/*
 * dump-metrics.c --
 *
 *      Operations to compute and dump MIB metrics.
 *
 * Copyright (c) 2000 Frank Strauss, Technical University of Braunschweig.
 * Copyright (c) 2000 J. Schoenwaelder, Technical University of Braunschweig.
 *
 * See the file "COPYING" for information on usage and redistribution
 * of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 *
 * @(#) $Id: dump-metrics.c,v 1.1 2000/05/20 12:51:11 strauss Exp $
 */

/*
  # revisions
  # imports
  # row creations:
  <basetype usage distribution>
  <complexity of table indexes>
 */

#include <config.h>

#include <stdio.h>
#include <string.h>

#include "smi.h"
#include "smidump.h"


typedef struct StatusCounter {
    unsigned int total;
    unsigned int current;
    unsigned int deprecated;
    unsigned int obsolete;
} StatusCounter;



typedef struct Metrics {
    StatusCounter tables;
    StatusCounter columns;
    StatusCounter scalars;
    StatusCounter notifications;
    unsigned int writable;
} Metrics;



typedef struct Module {
    char          *moduleName;
    struct Module *nextPtr;
} Module;

static Module *moduleList = NULL;



static void addModule(char *name)
{
    Module *mPtr;

    mPtr = xmalloc(sizeof(Module));
    mPtr->moduleName = xstrdup(name);
    mPtr->nextPtr = moduleList;
    moduleList = mPtr;
}



static void freeModules()
{
    Module *mPtr;

    while (moduleList) {
	mPtr = moduleList;
	moduleList = moduleList->nextPtr;
	xfree(mPtr->moduleName);
	xfree(mPtr);
    }
}



static void incrStatusCounter(StatusCounter *cnt, SmiStatus smiStatus)
{
    cnt->total++;
    switch (smiStatus) {
    case SMI_STATUS_CURRENT:
    case SMI_STATUS_MANDATORY:
    case SMI_STATUS_OPTIONAL:
	cnt->current++;
	break;
    case SMI_STATUS_DEPRECATED:
	cnt->deprecated++;
	break;
    case SMI_STATUS_OBSOLETE:
	cnt->obsolete++;
	break;
    case SMI_STATUS_UNKNOWN:
	break;
    }
}



static void addMetrics(Metrics *metrics, SmiModule *smiModule)
{
    SmiNode *smiNode;

    for (smiNode = smiGetFirstNode(smiModule, SMI_NODEKIND_ANY);
	 smiNode;
	 smiNode = smiGetNextNode(smiNode, SMI_NODEKIND_ANY)) {
	switch (smiNode->nodekind) {
	case SMI_NODEKIND_TABLE:
	    incrStatusCounter(&metrics->tables, smiNode->status);
	    break;
	case SMI_NODEKIND_COLUMN:
	    incrStatusCounter(&metrics->columns, smiNode->status);
	    if (smiNode->access == SMI_ACCESS_READ_WRITE) {
		metrics->writable++;
	    }
	    break;
	case SMI_NODEKIND_SCALAR:
	    incrStatusCounter(&metrics->scalars, smiNode->status);
	    if (smiNode->access == SMI_ACCESS_READ_WRITE) {
		metrics->writable++;
	    }
	    break;
	case SMI_NODEKIND_NOTIFICATION:
	    incrStatusCounter(&metrics->notifications, smiNode->status);
	    break;
	}
    }
}


static void printStatusCounter(StatusCounter *cnt, char *s)
{
    printf("%-14s %3d (%3d current, %3d deprecated, %3d obsolete)\n", s,
	   cnt->total, cnt->current, cnt->deprecated, cnt->obsolete);
}


static void printMetrics(Metrics *metrics)
{
    unsigned int objects;

    objects = metrics->columns.total + metrics->scalars.total;
    
    printStatusCounter(&metrics->tables, "Tables:");
    printStatusCounter(&metrics->columns, "Columns:");
    printStatusCounter(&metrics->scalars, "Scalars:");
    printStatusCounter(&metrics->notifications, "Notifications:");

    printf("%-14s %3d (%d%%)\n", "Writable:", metrics->writable,
	   objects ? (metrics->writable * 100 / objects) : 0);
}



int dumpMetrics(char *modulename, int flags)
{
    SmiModule *smiModule;
    Module    *mPtr;
    Metrics   metrics;
    int       unite;

    unite = flags & SMIDUMP_FLAG_UNITE;

    if (modulename) {
	addModule(modulename);
    }

    if ((modulename && !unite) || (!modulename && unite)) {
	if (! (flags & SMIDUMP_FLAG_SILENT)) {
	    printf("# %s module metrics (generated by smidump "
		   VERSION ")\n\n", unite ? "united" : modulename);
	}

	memset(&metrics, 0, sizeof(Metrics));
	if (modulename) {
	    smiModule = smiGetModule(modulename);
	    if (! smiModule) {
		fprintf(stderr, "smidump: cannot locate module `%s'\n",
			modulename);
		exit(1);
	    }
	    addMetrics(&metrics, smiModule);
	} else {
	    for (mPtr = moduleList; mPtr; mPtr = mPtr->nextPtr) {
		smiModule = smiGetModule(mPtr->moduleName);
		if (! smiModule) {
		    fprintf(stderr, "smidump: cannot locate module `%s'\n",
			    mPtr->moduleName);
		    exit(1);
		}
		addMetrics(&metrics, smiModule);
	    }
	}

	printMetrics(&metrics);
	
	freeModules();
    }

    return 0;
}
